/*
 * Copyright (c) 2008 Jiri Svoboda
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup rtld rtld
 * @brief
 * @{
 */ 
/**
 * @file
 */

#include <stdio.h>

#include <arch.h>
#include <elf_dyn.h>
#include <symbol.h>
#include <rtld.h>

void module_process_pre_arch(module_t *m)
{
	/* Unused */
}


/**
 * Process (fixup) all relocations in a relocation table.
 */
void rel_table_process(module_t *m, elf_rel_t *rt, size_t rt_size)
{
	int i;

	size_t rt_entries;
	size_t r_offset;
	elf_word r_info;
	unsigned rel_type;
	elf_word sym_idx;
	uintptr_t sym_addr;
	
	elf_symbol_t *sym_table;
	elf_symbol_t *sym;
	uint32_t *r_ptr;
	char *str_tab;
	
	elf_symbol_t *sym_def;
	module_t *dest;

	DPRINTF("parse relocation table\n");

	sym_table = m->dyn.sym_tab;
	rt_entries = rt_size / sizeof(elf_rel_t);
	str_tab = m->dyn.str_tab;

	DPRINTF("address: 0x%x, entries: %d\n", (uintptr_t)rt, rt_entries);
	
	for (i = 0; i < rt_entries; ++i) {
//		DPRINTF("symbol %d: ", i);
		r_offset = rt[i].r_offset;
		r_info = rt[i].r_info;

		sym_idx = ELF32_R_SYM(r_info);
		sym = &sym_table[sym_idx];

/*		DPRINTF("name '%s', value 0x%x, size 0x%x\n",
		    str_tab + sym->st_name,
		    sym->st_value,
		    sym->st_size);
*/
		rel_type = ELF32_R_TYPE(r_info);
		r_ptr = (uint32_t *)(r_offset + m->bias);

		if (sym->st_name != 0) {
//			DPRINTF("rel_type: %x, rel_offset: 0x%x\n", rel_type, r_offset);
			sym_def = symbol_def_find(str_tab + sym->st_name,
			    m, &dest);
//			DPRINTF("dest name: '%s'\n", dest->dyn.soname);
//			DPRINTF("dest bias: 0x%x\n", dest->bias);
			if (sym_def) {
				sym_addr = symbol_get_addr(sym_def, dest);
//				DPRINTF("symbol definition found, addr=0x%x\n", sym_addr);
			} else {
				DPRINTF("symbol definition not found\n");
				continue;
			}
		}

		switch (rel_type) {
		case R_386_GLOB_DAT:
		case R_386_JUMP_SLOT:
//			DPRINTF("fixup R_386_GLOB_DAT/JUMP_SLOT (b+v)\n");
			*r_ptr = sym_addr;
			break;

		case R_386_32:
//			DPRINTF("fixup R_386_32 (b+v+a)\n");
			*r_ptr += sym_addr;
			break;
			
		case R_386_RELATIVE:
//			DPRINTF("fixup R_386_RELATIVE (b+a)\n");
			*r_ptr += m->bias;
			break;

		default:
			printf("Unknown relocation type %d\n", rel_type);
			break;
		}

	}

}

void rela_table_process(module_t *m, elf_rela_t *rt, size_t rt_size)
{
	/* Unused */
	(void)m; (void)rt; (void)rt_size;
}

/** @}
 */
